/***************************************************************************
 *   Copyright (C) 2007 by Alexis Ménard <darktears31@gmail.com>           *
 *   Copyright 2009 by Giulio Camuffo <giuliocamuffo@gmail.com>		   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA .        *
 ***************************************************************************/

//own
#include "devicemanager.h"
#include "managerview.h"
#include "managerdialog.h"

//Qt
#include <QGraphicsLinearLayout>
#include <QGraphicsProxyWidget>
#include <QTimer>

//KDE
#include <KIcon>
#include <KConfigDialog>
#include <KStandardDirs>
#include <KDesktopFile>
#include <kdesktopfileactions.h>
#include <KIconLoader>
#include <KIO/Job>

//plasma
#include <Plasma/Dialog>
//use for desktop view
#include <Plasma/IconWidget>
#include <Plasma/Theme>

//solid
#include <solid/device.h>
#include <solid/storagedrive.h>
#include <solid/opticaldisc.h>
#include <solid/storageaccess.h>
#include <solid/opticaldrive.h>

using namespace Plasma;
using namespace Manager;

K_EXPORT_PLASMA_APPLET(devicemanager, DeviceManager)

DeviceManager::DeviceManager(QObject *parent, const QVariantList &args)
    : Plasma::PopupApplet(parent, args),
      m_solidEngine(0),
      m_solidDeviceEngine(0),
      m_icon(0),
      m_iconName(""),
      m_dialog(0),
      m_numberItems(0),
      m_itemsValidity(0),
      m_showAll(false)
{
    setBackgroundHints(StandardBackground);
    setAspectRatioMode(IgnoreAspectRatio);

    // let's initialize the widget
    resize(widget()->sizeHint());
}

DeviceManager::~DeviceManager()
{
    delete m_icon;
    delete m_dialog;
    delete m_popupTimer;
}

void DeviceManager::init()
{
    KConfigGroup cg = config();
    m_numberItems = cg.readEntry("NumberItems", 4);
    m_itemsValidity = cg.readEntry("ItemsValidity", 5);
    m_automount = cg.readEntry("automount", true);
    m_hidePopupAfter = cg.readEntry("hidePopupAfter", 5);
    m_showOnlyRemovable = cg.readEntry("showOnlyRemovable", false);
    m_showPopupOnInsert = cg.readEntry("showPopupOnInsert", true);
    m_layActionsInColumn = cg.readEntry("layActionsInColumn", true);

    m_solidEngine = dataEngine("hotplug");
    m_solidDeviceEngine = dataEngine("soliddevice");

    m_icon = new Plasma::IconWidget(KIcon("device-notifier", NULL), QString());
    m_iconName = QString("device-notifier");
    m_popupTimer = new QTimer();
    m_popupTimer->setSingleShot(true);
    connect(m_popupTimer, SIGNAL(timeout()), this, SLOT(hidePopup()));
    connect(m_dialog, SIGNAL(deviceSelected()), m_popupTimer, SLOT(stop()));

    Plasma::ToolTipManager::self()->registerWidget(this);

    setPopupIcon(m_icon->icon());

    m_dialog->setActionsLayoutInColumn(m_layActionsInColumn);

    //feed the list with what is already reported by the engine

    //connect to engine when a device is plug
    connect(m_solidEngine, SIGNAL(sourceAdded(const QString&)),
            this, SLOT(onSourceAdded(const QString&)));
    connect(m_solidEngine, SIGNAL(sourceRemoved(const QString&)),
            this, SLOT(onSourceRemoved(const QString&)));

    fillPreviousDevices();

}

QWidget *DeviceManager::widget()
{
    if (!m_dialog) {
        m_dialog = new ManagerDialog(this);
        connect(m_dialog, SIGNAL(actionSelected()), this, SLOT(hidePopup()));
        connect(m_dialog, SIGNAL(mountDevice(const QString&)), this, SLOT(mountDevice(const QString&)));
    }

    return m_dialog->dialog();
}

void DeviceManager::fillPreviousDevices()
{
    m_fillingPreviousDevices = true;
    foreach (const QString &source, m_solidEngine->sources()) {
        Solid::Device device = Solid::Device(source);
        Solid::Device parentDevice = device.parent();
        if (m_showOnlyRemovable) {
            Solid::StorageDrive *drive = parentDevice.as<Solid::StorageDrive>();
            if (drive && (drive->isHotpluggable() || drive->isRemovable()))
                onSourceAdded(source);
        }
        else
            onSourceAdded(source);
    }
    m_fillingPreviousDevices = false;
}

void DeviceManager::changeManagerIcon(const QString& name)
{
    if (m_icon && name.isNull()) {
        m_icon->setIcon(m_iconName);
    } else if (m_icon) {
        m_icon->setIcon(name);
    }

    setPopupIcon(m_icon->icon());
}

void DeviceManager::popupEvent(bool show)
{
    if (show) {
        Plasma::ToolTipManager::self()->clearContent(this);
    }
    else
	m_dialog->removeActions();
}

void DeviceManager::dataUpdated(const QString &source, Plasma::DataEngine::Data data)
{
    if (data.size() > 0) {
        //data from hotplug engine
        if (!data["predicateFiles"].isNull()) {
            int nb_actions = 0;
            QString lastActionLabel;
            foreach (const QString &desktop, data["predicateFiles"].toStringList()) {
                QString filePath = KStandardDirs::locate("data", "solid/actions/" + desktop);
                QList<KServiceAction> services = KDesktopFileActions::userDefinedServices(filePath, true);
                nb_actions += services.size();
                if (services.size() > 0) {
                    lastActionLabel = QString(services[0].text());
                }
            }
            m_dialog->setDeviceData(source,data["predicateFiles"],ManagerDialog::PredicateFilesRole);
            m_dialog->setDeviceData(source,data["text"], Qt::DisplayRole);

            //icon name
            m_dialog->setDeviceData(source,data["icon"], ManagerDialog::IconNameRole);
            //icon data
            m_dialog->setDeviceData(source,KIcon(data["icon"].toString()), Qt::DecorationRole);

            if (nb_actions > 1) {
                QString s = i18np("1 action for this device",
                                  "%1 actions for this device",
                                  nb_actions);
                m_dialog->setDeviceData(source, s, ManagerDialog::ActionRole);
            } else {
                m_dialog->setDeviceData(source, lastActionLabel, ManagerDialog::ActionRole);
            }
        //data from soliddevice engine
        } else {
            kDebug() << "DeviceManager::solidDeviceEngine updated" << source;
            if (data["Device Types"].toStringList().contains("Storage Access")) {
                QList<QVariant> freeSpaceData;
                freeSpaceData << QVariant(0) << QVariant(0);
                m_dialog->setDeviceData(source, QVariant(freeSpaceData), ManagerDialog::DeviceFreeSpaceRole);

                if (data["Accessible"].toBool() == true) {
                    m_dialog->setUnMount(true,source);

                    //set icon to mounted device
                    QStringList overlays;
                    overlays << "emblem-mounted";
                    m_dialog->setDeviceData(source, KIcon(m_dialog->getDeviceData(source, ManagerDialog::IconNameRole).toString(), NULL, overlays), Qt::DecorationRole);
		    
		    kDebug()<<"Free Space is valid: "<<data["Free Space"].isValid();
		    if (data["Free Space"].isValid()) {
                        QList<QVariant> freeSpaceData;
                        freeSpaceData << data["Size"] << data["Free Space"];
                        m_dialog->setDeviceData(source, QVariant(freeSpaceData), ManagerDialog::DeviceFreeSpaceRole);
                    }
		} else if (data["Device Types"].toStringList().contains("OpticalDisc")) {
                    //Unmounted optical drive
                    m_dialog->setDeviceData(source, KIcon("media-eject"), Qt::DecorationRole);
                    //set icon to unmounted device
                    m_dialog->setUnMount(false,source);
                    m_dialog->setDeviceData(source, KIcon(m_dialog->getDeviceData(source, ManagerDialog::IconNameRole).toString()), Qt::DecorationRole);
                } else {
                    m_dialog->setUnMount(false,source);

                    //set icon to unmounted device
                    m_dialog->setDeviceData(source, KIcon(m_dialog->getDeviceData(source, ManagerDialog::IconNameRole).toString()), Qt::DecorationRole);
                }
            }
            // actions specific for other types of devices will go here
        }
   }
}

void DeviceManager::notifyDevice(const QString &name)
{
    m_lastPlugged<<name;

    if (!m_fillingPreviousDevices) {
	if (m_showPopupOnInsert) {
	    showPopup();
	    if (m_hidePopupAfter != 0)
		hidePopupAfter(m_hidePopupAfter);
	}
	else {
	    changeManagerIcon("preferences-desktop-notification");
	    update();
	    QTimer::singleShot(5000, this, SLOT(changeManagerIcon()));
	}
    }
}

void DeviceManager::toolTipAboutToShow()
{
    Plasma::ToolTipContent toolTip;
    if (!m_lastPlugged.isEmpty()) {
        Solid::Device device(m_lastPlugged.last());

        toolTip.setSubText(i18n("Last plugged in device: %1", device.product()));
        toolTip.setImage(KIcon(device.icon()));
    } else {
        toolTip.setSubText(i18n("No devices plugged in"));
        toolTip.setImage(KIcon("device-notifier"));
    }

    Plasma::ToolTipManager::self()->setContent(this, toolTip);
}

void DeviceManager::toolTipHidden()
{
    Plasma::ToolTipManager::self()->clearContent(this);
}

void DeviceManager::removeLastDeviceNotification(const QString &name)
{
    m_lastPlugged.removeAll(name);
}

void DeviceManager::onSourceAdded(const QString &name)
{
    kDebug() << "DeviceManager:: source added" << name;
    if (m_dialog->countDevices() == m_numberItems && m_numberItems != 0) {
        QString itemUdi = m_dialog->getDeviceUdi(m_dialog->countDevices() - 1);
        //disconnect sources and after (important) remove the row
        m_solidDeviceEngine->disconnectSource(itemUdi, this);
        m_solidEngine->disconnectSource(itemUdi, this);

        m_dialog->removeDevice(m_dialog->countDevices() - 1);
    }

    KConfigGroup cg = config();
    bool visibility = cg.readEntry(name, true);

    if (visibility or m_showAll) {
        m_dialog->insertDevice(name);
        m_dialog->setDeviceData(name, visibility, ManagerDialog::VisibilityRole);
        notifyDevice(name);

        m_solidEngine->connectSource(name, this);
        m_solidDeviceEngine->connectSource(name, this);

        if (m_automount)
            m_dialog->mountDevice(name);
    }
}

void DeviceManager::onSourceRemoved(const QString &name)
{
    m_solidEngine->disconnectSource(name, this);
    m_solidDeviceEngine->disconnectSource(name, this);

    m_dialog->removeDevice(name);
    removeLastDeviceNotification(name);
}

void DeviceManager::resetDevices() {
    while (m_lastPlugged.count() > 0) {
	QString name = m_lastPlugged.takeAt(0);
	onSourceRemoved(name);
    }
    
    fillPreviousDevices();
}

void DeviceManager::createConfigurationInterface(KConfigDialog *parent) {
    QWidget *configurationWidget = new QWidget();
    m_configurationUi.setupUi(configurationWidget);
    parent->addPage(configurationWidget, i18n("General"), icon());
    parent->setButtons( KDialog::Ok | KDialog::Cancel);
    connect(parent, SIGNAL(okClicked()), this, SLOT(configAccepted()));

    m_configurationUi.automount->setChecked(m_automount);
    m_configurationUi.hidePopupAfter->setValue(m_hidePopupAfter);
    m_configurationUi.showPopupOnInsert->setChecked(m_showPopupOnInsert);
    m_configurationUi.showOnlyRemovable->setChecked(m_showOnlyRemovable);
    m_configurationUi.layActionsInColumn->setChecked(m_layActionsInColumn);
    m_configurationUi.layActionsInRow->setChecked(!m_layActionsInColumn);
}

void DeviceManager::configAccepted() {
    KConfigGroup cg = config();

    m_automount = m_configurationUi.automount->isChecked();
    m_hidePopupAfter = m_configurationUi.hidePopupAfter->value();
    m_showPopupOnInsert = m_configurationUi.showPopupOnInsert->isChecked();
    m_showOnlyRemovable = m_configurationUi.showOnlyRemovable->isChecked();
    m_layActionsInColumn = m_configurationUi.layActionsInColumn->isChecked();
    m_dialog->setActionsLayoutInColumn(m_layActionsInColumn);

    cg.writeEntry("automount", m_automount);
    cg.writeEntry("hidePopupAfter", m_hidePopupAfter);
    cg.writeEntry("showPopupOnInsert", m_showPopupOnInsert);
    cg.writeEntry("showOnlyRemovable", m_showOnlyRemovable);
    cg.writeEntry("layActionsInColumn", m_layActionsInColumn);

    resetDevices();
}

void DeviceManager::hidePopupAfter(int seconds) {
    m_popupTimer->start(seconds * 1000);
}

void DeviceManager::setItemShown(const QString &name, bool shown) {

    m_dialog->setDeviceData(name, shown, ManagerDialog::VisibilityRole);
    if (!shown and !m_showAll)
        onSourceRemoved(name);

    KConfigGroup cg = config();
    cg.writeEntry(name, shown);
}

void DeviceManager::setAllItemsShown(bool shown) {
    m_showAll = shown;
    resetDevices();
}
	
#include "devicemanager.moc"
